module PolynomialSystems_v01
# parameterized polynomial systems

export Newton_Raphson, get_hessian, get_residual
export lexmax, readSystem

using MultivariatePolynomials
# using DynamicPolynomials
using TypedPolynomials
import LinearAlgebra: norm

# to allow substitutions on arrays of polynomials
(a::Array)(x...) = map(y->y(x...), a)

# Return a matrix with all exponent tuples in n variables of degree at most deg
function lexmax(n, deg)
    if n == 1
            M = collect(0:deg);
    else
            N = binomial(deg+n,n);
            M = fill(0,n,N);
            index = 0;
            for k = 0:deg
                Mtemp = lexmax(n-1,deg-k);
                if n == 2
                    ntemp = length(Mtemp)
                    Mtemp = Mtemp'
                else
                    ntemp = size(Mtemp,2);
                end
                M[:,index+1:index+ntemp] = [ fill(1,1,ntemp)*k; Mtemp];
                index = index + ntemp;
            end
    end
    return M
end

# compute a solution using Newton-Raphson for a square pol system
function Newton_Raphson(polsys,J, x, sol0::Array{Complex{Float64},1}, maxnsteps::Int)
    sol = sol0
    i = 1
    f = polsys(x => sol)
    dsol = -(J(x => sol) \ f)
    while norm(dsol)/(norm(sol)+1) > 10^2*eps() && i < maxnsteps    #  nsteps of Newton-Raphson
        #println("value of sol: $sol")
        #println("value of f: $f")
        sol = sol + dsol
        f = polsys(x => sol)
        dsol = -(J(x => sol) \ f)
        i = i+1
        #=
        if i == maxnsteps
            sleep(10)
        end
        =#
    end
    #println("Newton: $i steps")
    return sol
end

# A function to compute the Hessian of an n-variate polynomial
function get_hessian(f,x)
        n = length(x)
        H = fill(f,n,n)
        for i = 1:n
                for j = 1:i
                        H[i,j] = differentiate(f,x[i])
                        H[i,j] = differentiate(H[i,j],x[j])
                        H[j,i] = H[i,j]
                end
        end
        return H
end

function get_residual(polsys, sol::Array{Array{Complex{Float64},1},1}, x)
        #println("COMPUTING RESIDUALS")
        #println("-------------------")
        #println("INPUT:")
        #println("$polsys")
        #println("$sol")
        #println("$x")
        n = length(x)
        residuals = zeros(length(sol))
        for k = 1:length(sol)
                ressol = 0
                for i = 1:length(polsys)
                        R = polsys[i]
                        T = terms(R)
                        l = length(T)
                        Rabs = maximum(abs.([T[s](x=>abs.(sol[k])) for s=1:l]))
                        R = R(x=>sol[k])/(1+Rabs)
                        ressol = ressol + abs(R)
                end
                residuals[k] = ressol/length(polsys)
        end
        #println("----------------------")
        #println("OUTPUT:")
        #println("residuals = $residuals")
        return residuals
end

function readStartSolutions(filename, x)
    # Read the start solutions from a file generated by Matlab
    SOLS = readlines(filename)
    delta = size(SOLS,1)
    n = length(x)
    startsol = fill(fill(0.0+0.0*im,n), delta)
    for i = 1:delta
        startsol[i] = eval(Meta.parse("[" * replace(SOLS[i],"  " => ", ") * "]"))
    end
    return startsol
end


# end of module
end
